from .vqleditor import VqlSyntaxHighlighter, VqlEditor  # deprecated

from .markdown_editor import MarkdownEditor

from .dict_widget import DictModel, DictWidget

from .codeedit import CodeEdit

from .file_edit import FileEdit
