from PySide2.QtWidgets import QApplication

__title__ = "VQL History"
__description__ = "A VQL History"
__long_description__ = QApplication.instance().translate(
    "vql_history",
    """
<p>This plugin store all previous VQL executed </p>


""",
)
__author__ = "Sacha schutz"
__version__ = "1.0.0"
